//#*#*#*#*#*#*#*#*#*#*#*#*#*#*#*#* DtMxAncPacketDemo.cpp *#*#*#*#*#*#*#*#* (C) 2017 DekTec
//
// This example application demonstrates a matrix configuration for recieving ANC packets.
// The application reports all anc packets it detects in the incoming stream.
// 

//.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.- Include files -.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-
#include <cstdio>
#include <cstring>
#ifdef _WIN32
    // Windows specific includes
    #include <windows.h>
    #include <conio.h>
#else
    // Linux specific includes
    #include <unistd.h>
    #include <limits.h>
    #include <termios.h>
    #include <sys/select.h>
#endif
#include "MxAncPacketDemo.h"

// All code resides in MxExamples namespace
using namespace MxExamples;

//-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.- Application Version -.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.
#define APP_NAME                "DtMxAncPacketDemo"
#define APP_VERSION_MAJOR       1
#define APP_VERSION_MINOR       0
#define APP_VERSION_BUGFIX      0

//.-.-.-.-.-.-.-.-.-.-.-.-.-.-.- Application configuration -.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-

// Default, card to attach to
static const int  DEF_CARD_TYPE = 2174;
// Default, number of card to attach to
static const int  DEF_CARD_NO = 0;
// Default, output port
static const int  DEF_IN_PORT=1;

// Default input video standard
#define DTAPI_VIDSTD_AUTO  -2   // Extra helper for auto detection of video standard
static const int  DEF_IN_VIDSTD = DTAPI_VIDSTD_AUTO;

// Row indices
static const int  IN_ROW=0;


// Enable this define to get profiling information
//#define  MX_PRINT_PROFILING_INFO

//=+=+=+=+=+=+=+=+=+=+=+=+=+=+=+=+=+=+=+ Main loop +=+=+=+=+=+=+=+=+=+=+=+=+=+=+=+=+=+=+=+

//-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.- main -.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-
//
int main(int  argc, char*  argv[])
{
    DTAPI_RESULT  dr;
    DtDevice  TheCard;
    MxAncPacketDemo  R;    // The recorder

#ifndef _WIN32
    ChangeTerminalMode(1);  // Set terminal mode, so that we can use _kbhit()
#endif

    try
    {
        //-.-.-.-.-.-.-.-.-.-.-.-.-.- Display welcome message -.-.-.-.-.-.-.-.-.-.-.-.-.-.

        printf("%s V%d.%d.%d (c) 2017 DekTec Digital Video B.V.\n\n", APP_NAME,
                                APP_VERSION_MAJOR, APP_VERSION_MINOR, APP_VERSION_BUGFIX);

        //.-.-.-.-.-.-.-.-.-.-.-.- Parse the command-line options -.-.-.-.-.-.-.-.-.-.-.-.

        R.ParseCommandLine(argc, argv);        

        //.-.-.-.-.-.-.-.-.-.-.-.-.-.-.- Attach to the card -.-.-.-.-.-.-.-.-.-.-.-.-.-.-.

        dr = TheCard.AttachToType(R.CARD_TYPE(), R.CARD_NO());
        if (dr != DTAPI_OK)
            MX_THROW_EXC(dr, "Failed to attach to DTA-%d", R.CARD_TYPE());

        // Check the port for CAP_MATRIX2. The DtMxXxx API is only supported by ports 
        // with DTAPI_CAP_MATRIX2 capability
        if ((TheCard.m_pHwf[(R.IN_PORT()-1)].m_Flags & DTAPI_CAP_MATRIX2) == 0)
            MX_THROW_EXC_NO_ERR("Port %d does not have the MATRIX2 capability", 
                                                                             R.IN_PORT());

        printf("  Using DTA-%d, Port=%d, Standard=%s\n", R.CARD_TYPE(),
                                               R.IN_PORT(), 
                                               R.IN_VIDSTD()==DTAPI_VIDSTD_AUTO ? 
                                               "auto" : ::DtapiVidStd2Str(R.IN_VIDSTD()));
                        
        //.-.-.-.-.-.-.-.-.-.-.-.-.-.- Create the PIP matrix -.-.-.-.-.-.-.-.-.-.-.-.-.-.-

        // Init the matrix
        if (!R.Init(TheCard))
            MX_THROW_EXC_NO_ERR("Failed to init demo");
        // Start the matrix
        if (!R.Start())
            MX_THROW_EXC_NO_ERR("Failed to start demo");

        //-.-.-.-.-.-.-.-.-.- Process loop, wait for user to press key -.-.-.-.-.-.-.-.-.-
        
        printf("\nPress any key to exit\n\n");
        while(!_kbhit()) { Sleep(10); }
        _getch();   // Read the keystroke from the console

        R.Stop(); // Stop the matrix

        // Print profling information
#ifdef MX_PRINT_PROFILING_INFO
        R.PrintProfilingInfo();
        printf("\nPress any key to exit\n");
        // Wait for a keystroke before exiting. To allow user time to take in the 
        // profiling information
        _getch();
#endif
    }
    catch(MxDemoExc  e)
    {
        puts(e.ErrorStr());
    }
    catch(...) 
    {
        puts("Encountered an unhandled exception!!!");
    }

#ifndef _WIN32
    ChangeTerminalMode(0);  // Restore terminal mode
#endif
    return 0;
}

//+=+=+=+=+=+=+=+=+=+=+=+=+=+ MxAncPacketDemo implementation +=+=+=+=+=+=+=+=+=+=+=+=+=+=

//.-.-.-.-.-.-.-.-.-.-.-.-.- MxAncPacketDemo::MxAncPacketDemo -.-.-.-.-.-.-.-.-.-.-.-.-.
//
MxAncPacketDemo::MxAncPacketDemo() 
{
}

//-.-.-.-.-.-.-.-.-.-.-.-.- MxAncPacketDemo::~MxAncPacketDemo -.-.-.-.-.-.-.-.-.-.-.-.-.
//
MxAncPacketDemo::~MxAncPacketDemo()
{
}

//.-.-.-.-.-.-.-.-.-.-.-.-.- MxAncPacketDemo::ParseCommandLine -.-.-.-.-.-.-.-.-.-.-.-.-.
//
void  MxAncPacketDemo::ParseCommandLine(int argc, char* argv[])
{
    // Table defining video standard command-line options
    static const DtEnumOptPair CMD_VIDSTD_OPTS[] = 
    {
        // Auto detected
        { L"auto",       DTAPI_VIDSTD_AUTO,         L"auto" },

        // 1080: Interlaced
        { L"1080i60",    DTAPI_VIDSTD_1080I60,      L"1080i/60" },
        { L"1080i59",    DTAPI_VIDSTD_1080I59_94,   L"1080i/59.94" },
        { L"1080i50",    DTAPI_VIDSTD_1080I50,      L"1080i/50" },
        
        // 1080: progressive
        { L"1080p60",    DTAPI_VIDSTD_1080P60,      L"1080p/60" },
        { L"1080p59",    DTAPI_VIDSTD_1080P59_94,   L"1080p/59.94" },
        { L"1080p50",    DTAPI_VIDSTD_1080P50,      L"1080p/50" },
        { L"1080p30",    DTAPI_VIDSTD_1080P30,      L"1080p/30" },
        { L"1080p29",    DTAPI_VIDSTD_1080P29_97,   L"1080p/29.97" },
        { L"1080p25",    DTAPI_VIDSTD_1080P25,      L"1080p/25" },
        { L"1080p24",    DTAPI_VIDSTD_1080P24,      L"1080p/24" },
        { L"1080p23",    DTAPI_VIDSTD_1080P23_98,   L"1080p/23.98" },

        // 720: progressive
        { L"720p60",     DTAPI_VIDSTD_720P60,       L"720p/60" },
        { L"720p59",     DTAPI_VIDSTD_720P59_94,    L"720p/59.94" },
        { L"720p50",     DTAPI_VIDSTD_720P50,       L"720p/50" },
        { L"720p30",     DTAPI_VIDSTD_720P30,       L"720p/30" },
        { L"720p29",     DTAPI_VIDSTD_720P29_97,    L"720p/29.97" },
        { L"720p25",     DTAPI_VIDSTD_720P25,       L"720p/25" },
        { L"720p24",     DTAPI_VIDSTD_720P24,       L"720p/24" },
        { L"720p23",     DTAPI_VIDSTD_720P23_98,    L"720p/23.98" },

        // SD: NTSC & PAL
        { L"525i59",     DTAPI_VIDSTD_525I59_94,    L"525i/59.94" },
        { L"625i50",     DTAPI_VIDSTD_625I50,       L"625i/50" },

        OPT_PAIR_END,
    };
    
    // Table defining the command-line options
    static DtOptItem  CMD_OPTIONS_DEF[] =
    {
        DtOptItem(L"t", m_OptCardType, DEF_CARD_TYPE, L"Card type", 0, 5000),
        DtOptItem(L"n", m_OptCardNo, DEF_CARD_NO, L"Card number", 0, 16),
        DtOptItem(L"p", m_OptInPort, DEF_IN_PORT, L"Input port", 1, 16),
        DtOptItem(L"v", m_OptInVidStd, DEF_IN_VIDSTD, L"Video standard", 
                                                                         CMD_VIDSTD_OPTS),
    };
    static const int  NUM_CMD_OPTIONS = sizeof(CMD_OPTIONS_DEF) / 
                                                               sizeof(CMD_OPTIONS_DEF[0]);
    try
    {
        // Actual parsing of the command line options
        std::list<std::wstring>  FreeArgs;
        DtOptItem::ParseOpt(CMD_OPTIONS_DEF, argc, argv, FreeArgs);
    }
    catch (DtOptException  e)
    {
        MX_THROW_EXC_NO_ERR("Command-line error: %ls", (const wchar_t*)e);
    }
}

//.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.- MxAncPacketDemo::Init -.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.
//
bool  MxAncPacketDemo::Init(DtDevice&  TheCard)
{
    DTAPI_RESULT  dr;

    //-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.- Prep card for matrix -.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-

    // The card must have been attached
    if (!TheCard.IsAttached())
        return false;
    if (!PrepCard(TheCard))
        return false;

    //+=+=+=+=+=+=+=+=+=+=+=+=+=+=+ Configure the input row +=+=+=+=+=+=+=+=+=+=+=+=+=+=+=

    // Step 1: Attach row to the input
    DtMxPort  InPort;
    dr = InPort.AddPhysicalPort(&TheCard, IN_PORT());
    if (dr != DTAPI_OK)
        MX_THROW_EXC(dr, "Failed to add port %d as physical in-port", IN_PORT());
    dr = m_TheMatrix.AttachRowToInput(IN_ROW, InPort);
    if (dr != DTAPI_OK)
        MX_THROW_EXC(dr, "Failed to attach row %d to in-port %d", IN_ROW, IN_PORT());

    // Step 2: configure the row
    DtMxRowConfig  RowConfig;
    RowConfig.m_Enable = true;
    RowConfig.m_RowSize = 1;       // Keep a history of one frame

    // Disable video+audio. Only interested in ANC packets
    RowConfig.m_VideoEnable = false;
    RowConfig.m_AudioEnable = false;
    
    // Enable AUX-data and mark ANC-packets for de-emdedding
    RowConfig.m_AuxDataEnable = true;
    RowConfig.m_AuxData.m_DataType = DT_AUXDATA_SDI;
    RowConfig.m_AuxData.m_Sdi.m_AncPackets.m_DeEmbed = true;

    // Apply the row configuration
    dr = m_TheMatrix.SetRowConfig(IN_ROW, RowConfig);
    if (dr != DTAPI_OK)
        MX_THROW_EXC(dr, "Failed set row %d config", IN_ROW);

    // Step 3: Final configuration

    // Register our callback
    dr = m_TheMatrix.AddMatrixCbFunc(MxAncPacketDemo::OnNewFrame, this);
    if (dr != DTAPI_OK)
        MX_THROW_EXC(dr, "Failed register callback function");

    return true;
}

//.-.-.-.-.-.-.-.-.-.-.-.-.-.-.- MxAncPacketDemo::PrepCard -.-.-.-.-.-.-.-.-.-.-.-.-.-.-.
//
bool  MxAncPacketDemo::PrepCard(DtDevice&  TheCard)
{
    DTAPI_RESULT  dr;
           
    //-.-.-.-.-.-.-.-.-.-.-.-.-.- Configure port io-direction -.-.-.-.-.-.-.-.-.-.-.-.-.-.

    DtIoConfig  Cfg;
    
    Cfg.m_Port = IN_PORT();
    Cfg.m_Group = DTAPI_IOCONFIG_IODIR;
    Cfg.m_Value = DTAPI_IOCONFIG_INPUT;
    Cfg.m_SubValue = DTAPI_IOCONFIG_INPUT;
    Cfg.m_ParXtra[0] = Cfg.m_ParXtra[1] = -1;
    
    dr = TheCard.SetIoConfig(&Cfg, 1);
    if (dr != DTAPI_OK)
        MX_THROW_EXC(dr, "Failed to apply IO-DIR for port");

    //-.-.-.-.-.-.-.-.-.-.-.-.-.- Configure port io-standard -.-.-.-.-.-.-.-.-.-.-.-.-.-.

    // Should we auto detect? 
    int  VidStd=IN_VIDSTD(), IoStdValue=-1, IoStdSubValue=-1;
    if (VidStd == DTAPI_VIDSTD_AUTO)
    {
        do
        {
            IoStdValue=-1, IoStdSubValue=-1;
            dr = TheCard.DetectIoStd(IN_PORT(), IoStdValue, IoStdSubValue);
            Sleep(10);

            if (_kbhit())
            {
                _getch();   // read keystroke from console
                return false;
            } 
        }
        while (dr != DTAPI_OK);

        ::DtapiIoStd2VidStd(IoStdValue, IoStdSubValue, VidStd);
        printf("  Detected video standard = %s\n", ::DtapiVidStd2Str(VidStd));
    }
    
    dr = ::DtapiVidStd2IoStd(VidStd, IoStdValue, IoStdSubValue);
    if (dr != DTAPI_OK)
        MX_THROW_EXC(dr, "Failed to deduce IO-STD from video and link standard");
    
    Cfg.m_Port = IN_PORT();
    Cfg.m_Group = DTAPI_IOCONFIG_IOSTD;
    Cfg.m_Value = IoStdValue;
    Cfg.m_SubValue = IoStdSubValue;
    Cfg.m_ParXtra[0] = Cfg.m_ParXtra[1] = -1;
    
    dr = TheCard.SetIoConfig(&Cfg, 1);
    if (dr != DTAPI_OK)
        MX_THROW_EXC(dr, "Failed to apply IO-STD for port");

    return (dr == DTAPI_OK);
}

//-.-.-.-.-.-.-.-.-.-.-.-.-.-.- MxAncPacketDemo::OnNewFrame -.-.-.-.-.-.-.-.-.-.-.-.-.-.-
//
void  MxAncPacketDemo::OnNewFrame(DtMxData* pData, void* pOpaque)
{
    ((MxAncPacketDemo*)pOpaque)->OnNewFrame(pData);
}

//-.-.-.-.-.-.-.-.-.-.-.-.-.-.- MxAncPacketDemo::OnNewFrame -.-.-.-.-.-.-.-.-.-.-.-.-.-.-
//
void  MxAncPacketDemo::OnNewFrame(DtMxData* pData)
{
    // Get the frame data and check it is valid
    DtMxFrame*  pTheFrame = pData->m_Rows[IN_ROW].m_CurFrame;
    if (pTheFrame->m_Status != DT_FRMSTATUS_OK)
    {
        printf("[Frame %lld] frame status (%d) is not ok => skipping frame\n", 
                                                     pData->m_Frame, pTheFrame->m_Status);
        return;
    }

    //-.-.-.-.-.-.-.-.-.-.-.-.-.- Get the ANC-packets we want -.-.-.-.-.-.-.-.-.-.-.-.-.-.

    DTAPI_RESULT  dr;

    const int  SCTE104_DID = 0x41;
    const int  SCTE104_SDID = 0x07;
    // SCTE-104 packets are located in VANC-space
    const int  ANC_SPACE = DTAPI_SDI_VANC;
    // Look in both the chrominance streams
    const int  ANC_STREAM = DTAPI_SDI_CHROM;

    // First determine the number of packets available
    int  NumPackets = 0;
    dr = pTheFrame->AncGetPacket(SCTE104_DID, SCTE104_SDID, NULL, NumPackets, 
                                                                   ANC_SPACE, ANC_STREAM);

    // Now get the packets for real
    if (NumPackets > 0)
    {
        DtMxAncPacket*  pAncPackets = new DtMxAncPacket[NumPackets];
        dr = pTheFrame->AncGetPacket(SCTE104_DID, SCTE104_SDID, pAncPackets, 
                                                       NumPackets, ANC_SPACE, ANC_STREAM);

        // Print size per packet
        printf("Found %d SCTE-104:\n", NumPackets);
        for (int i=0; i<NumPackets; i++)
        {
            printf("- Packet %d: DID=0x%02X, SDID=0x%02X, DC=%d, Line=%d\n", i+1, 
                                         pAncPackets[i].m_Did, pAncPackets[i].m_SdidOrDbn,
                                         pAncPackets[i].m_Dc, pAncPackets[i].m_Line);
        }
        delete [] pAncPackets;
    }
    else
        printf("No SCTE-104 packets found\n");
}
