//*#*#*#*#*#*#*#*#*#*#*#*#*#*#*#*#*#* MxHdmiHdrDemo.h *#*#*#*#*#*#*#*#*#*# (C) 2018 DekTec
//

#ifndef  __MXHDMIHDRDEMO_H
#define  __MXHDMIHDRDEMO_H

//.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.- Include files -.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-
#include <string>
#include <math.h>
#include "DTAPI.h"
#include "MxExamplesCommon.h"

// All code resides in MxExamples namespace
namespace MxExamples
{

//-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.- class MxHdmiHdrDemo -.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.
//
class MxHdmiHdrDemo : public MxDemoMatrixBase
{
    // Types / Contants
public:
    //-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.- CallbackContext -.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.
    // Context passed to callback function
    struct CallbackContext
    {
        MxHdmiHdrDemo*  m_pPlayer; // Pointer back to the player
        enum
        {
            GEN_VIDEO,
            GEN_AUDIO,
            GEN_ANC,
        } m_Type;               // Indicates the type off callback to call
    };
    
    // Operations
public:
    virtual bool  Init(DtDevice&  TheCard);
protected:
    bool  PrepCard(DtDevice&  TheCard);
    static void  OnNewFrame(DtMxData* pData, void* pOpaque);
    void  OnGenerateVideo(DtMxData* pData);
    void  OnGenerateAudio(DtMxData* pData);
    void  OnGenerateAnc(DtMxData* pData);

    // Data / Attributes
public:
protected:
    CallbackContext  m_ContextAudio;  // Context object for audio callback
    CallbackContext  m_ContextVideo;  // Context object for video callback
    CallbackContext  m_ContextAnc;    // Context object for anc callback
    
    MxRectangle  m_BoundingRect;        // Bounding MxRectangle for the image
    std::vector<MxRectangle>  m_Rectangles;  // Floating rectangles
    void  InitChannelStatus(DtMxAudioChannel& Channel, int AudioFs);
    DtAvOutput  m_AvOutput;
    DtMxRowConfig  m_RowConfig;
    
    // Constructor / Destructor
public:
    MxHdmiHdrDemo();
    virtual ~MxHdmiHdrDemo();
};

//.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.- MxToneGeneratorEx -.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-
// Extends the standard tone generator with functionaity to generate AES samples
class MxToneGeneratorEx : protected MxToneGenerator
{
    // Constants / Types
public:
    static const int MAX_NUM_CHANNELS=4;
    static const bool PARITY_TABLE256[256];     // Parity lookup table

    // Operations
public:
    void Init();
    void  Generate(unsigned int*  pSamples,int  NumSamples2Gen);
protected:
    void  GenerateAes3(unsigned int*  pSamples,int  NumSamples2Gen);
    bool  Parity32(unsigned int  Data);
    
    // Data / Attributes
public:
protected:
    DtMxAudioSampleType  m_AudioSampleFormat;
    const int  m_Channel;       // Audio channel number
    int  m_NextStatusWordBit[MAX_NUM_CHANNELS];   // Next status word bit to insert
private:
    bool  m_StatusWordIsValid[MAX_NUM_CHANNELS];  // true, is status wors has been initialised
    unsigned char  m_StatusWord[MAX_NUM_CHANNELS][24];  // Buffer with audio status word
    
    // Constructor / Destructor
public:
    MxToneGeneratorEx(const DtMxAudioSampleType AudioType, int  Channel, int  ToneFreq, double  Attn, int Fs=48000);
    virtual ~MxToneGeneratorEx();
};

//-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.- MxAudioStatusWordParser -.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.
//
// Provides access function to the bit-fields of the audio status word. Refer to 
// IEC-60958-4 (2008) specification for details on each of the bit(-field)s
//
class MxAudioStatusWordParser
{
    // Constants / Types
public:
    static const int  SIZE_NUM_BITS = 192;
    static const unsigned char  BIT_MASK[9];
    
    // Operations
public:
    void  Attach(unsigned char  Data[24]);

#define GET_SET_FUNC(Name, BitNr, NumBits)                                               \
    inline Mx##Name  Get##Name() { return (Mx##Name)GetBits(BitNr, NumBits); }           \
    inline void  Set##Name(Mx##Name  Value) { SetBits((unsigned char)Value, BitNr, NumBits); }
    
    // Use of channel status block: bit0
    enum  MxProfessionalUse
    {
        MX_USE_COMMERCIAL = 0,      // Commercial use of channel status block
        MX_USE_PROFESSIONAL = 1     // Professional use of channel status block
    };
    GET_SET_FUNC(ProfessionalUse, 0, 1);
    // Linear PCM identification: bit1
    enum  MxLinearPcm
    {
        MX_LINEARPCM_PCM = 0,       // Audio sample word represents linear PCM samples
        MX_LINEARPCM_DATA = 1       // Audio sample word used for purposes other than
                                    // linear PCM samples
    };
    GET_SET_FUNC(LinearPcm, 1, 1);
    // Encoded audio signal pre-emphasis: bit2..4
    enum MxPreEmphasis
    {
        MX_PREEMPHASIS_NOTSET = 0,  // Pre-emphasis not indicated. Receiver defaults to 
                                     // no pre-emphasis with manual override enabled
        MX_PREEMPHASIS_RESV1 = 1,
        MX_PREEMPHASIS_RESV2 = 2,
        MX_PREEMPHASIS_RESV3 = 3,
        MX_PREEMPHASIS_NONE = 4,     // No pre-emphasis
        MX_PREEMPHASIS_RESV5 = 5,
        MX_PREEMPHASIS_50_15 = 6,    // 50 us/15 us pre-emphasis
        MX_PREEMPHASIS_ITU_J17 = 7,  // ITU-T Recommendation J.17 pre-emphasis (with
                                     // 6,5 dB insertion loss at 800 Hz)
    };
    GET_SET_FUNC(PreEmphasis, 2, 3);
    // Lock indication: bit5
    enum MxLockIndication
    {
        MX_LOCK_DEFAULT = 0,        // Default, lock condition not indicated
        MX_LOCK_UNLOCKED = 1,       // Source sampling frequency unlocked
    };
    GET_SET_FUNC(LockIndication, 5, 1);
    // Encoded sampling frequency: bit6..7
    enum MxEncodedSamplingFreq
    {
        MX_ENCSAMPLFREQ_NOTSET = 0, // Sampling frequency not indicated. Receiver defaults
                                    // to 48 kHz and manual override or auto set is 
                                    // enabled
        MX_ENCSAMPLFREQ_48000 = 1,  // 48 kHz sampling frequency
        MX_ENCSAMPLFREQ_44100 = 2,  // 44.1 kHz sampling frequency
        MX_ENCSAMPLFREQ_32000 = 3   // 32 kHz sampling frequency
    };
    GET_SET_FUNC(EncodedSamplingFreq, 6, 2);
    int  GetEncodedSamplingFreqHz();
    void  SetEncodedSamplingFreqHz(int  FreqHz);

    // Encoded channel mode: bit8..11
    enum MxChannelMode
    {
        MX_CHANMODE_UNSPEC = 0,     // Unspecified mode
        MX_CHANMODE_TWOCHAN = 1,    // Two-channel mode
        MX_CHANMODE_MONO = 2,       // Single-channel mode
        MX_CHANMODE_PRISEC = 3,     // Primairy/Secondary mode
        MX_CHANMODE_STEREO = 4,     // Stereo mode
        MX_CHANMODE_RESV1 = 5,      // Reserved for future use
        MX_CHANMODE_RESV2 = 6,      // Reserved for future use
        MX_CHANMODE_DBLRATE = 7,    // Single channel double-sampling rate mode
        MX_CHANMODE_DBLRATE_L = 8,  // Stereophonic left at double-sampling rate  mode
        MX_CHANMODE_DBLRATE_R = 9,  // Stereophonic right at double-sampling rate mode
        MX_CHANMODE_RESV3 = 10,     // Reserved for future use
        MX_CHANMODE_RESV4 = 11,     // Reserved for future use
        MX_CHANMODE_RESV5 = 12,     // Reserved for future use
        MX_CHANMODE_RESV6 = 13,     // Reserved for future use
        MX_CHANMODE_RESV7 = 14,     // Reserved for future use
        MX_CHANMODE_MULTI = 15      // Multi-channel mode
    };
    GET_SET_FUNC(ChannelMode, 8, 4);
    // Encoded user bits management: bit12..15
    enum MxUserBitsManagment
    {
        MX_USRBITSMAN_UNDEF = 0,    // Default, user data format is undefined
        MX_USRBITSMAN_192B = 1,     // 192-bit block structure. Preamble B indicates 
                                    // the start of the block
        MX_USRBITSMAN_AES18 = 2,    // Reserved for the AES18 standard
        MX_USRBITSMAN_USRDEF = 3,   // User-defined
        MX_USRBITSMAN_IEC60958_3 = 4,   // User data conforms to the general user data 
                                        // format as defined in IEC 60958-3
        MX_USRBITSMAN_AES52 = 5,    // Reserved for metadata as described in AES52
        MX_USRBITSMAN_RESV1 = 6,
        MX_USRBITSMAN_RESV2 = 7,
        MX_USRBITSMAN_RESV3 = 8,
        MX_USRBITSMAN_RESV4 = 9,
        MX_USRBITSMAN_RESV5 = 10,
        MX_USRBITSMAN_RESV6 = 11,
        MX_USRBITSMAN_RESV7 = 12,
        MX_USRBITSMAN_RESV8 = 13,
        MX_USRBITSMAN_RESV9 = 14,
        MX_USRBITSMAN_RESV10 = 15
    };
    GET_SET_FUNC(UserBitsManagment, 12, 4);
    // Encoded use of auxiliary sample bits: bit16..18
    enum MxUseOfAuxSampleBits
    {
        MX_AUXBITS_NOT_USED = 0,    // Max 20-bit audio samples, aux bits not used
        MX_AUXBITS_24_BIT = 1,      // Max 24-bit audio samples, aux bits used
        MX_AUXBITS_20_BIT = 2,      // Max 20-bit audio samples, aux bits used for single 
                                    // coordination signal 
        MX_AUXBITS_RESV1 = 3,       // Reserved for future use
    };
    GET_SET_FUNC(UseOfAuxSampleBits, 16, 3);
    // Encoded audio sample word length of transmitted signal: bit19..21
    enum MxAudioSampleWordLength
    {
        MX_SAMPLELENGTH_DEFAULT = 0, // Word length not indicated (use default=20)
        MX_SAMPLELENGTH_MIN1 = 1,    // Size indicated by AuxSampleBitsUse minus 1
        MX_SAMPLELENGTH_MIN2 = 2,    // Size indicated by AuxSampleBitsUse minus 2
        MX_SAMPLELENGTH_MIN3 = 3,    // Size indicated by AuxSampleBitsUse minus 3
        MX_SAMPLELENGTH_MIN4 = 4,    // Size indicated by AuxSampleBitsUse minus 4
        MX_SAMPLELENGTH_MAX = 5,     // Size indicated as by AuxSampleBitsUse
    };
    GET_SET_FUNC(AudioSampleWordLength, 19, 3);
    // Indication of alignment level: bit22..23
    enum MxAlignmentLevel
    {
        MX_ALIGNLVL_DEFAULT = 0,     // Alignment level not indicated
        MX_ALIGNLVL_SMPTE_RP155 = 1, // Alignment level is 20 dB below maximum code 
                                     // (refer to SMPTE RP155)
        MX_ALIGNLVL_EBU_R68 = 2,     // Alignment level is 18,06 dB below maximum code 
                                     // (refer to EBU R68)
        MX_ALIGNLVL_RESV1 = 3,       // Alignment level not indicated
    };
    GET_SET_FUNC(AlignmentLevel, 22, 2);

    // Multichannel mode control bit: bit31
    //GET_SET_FUNC(MultichannelModeControl, 31, 1);

    // Digital audio reference signal: bit32..33
    enum MxReferenceSignal
    {
        MX_REFSIG_NONE = 0,         // Not a reference signal (default).
        MX_REFSIG_GRADE1 = 1,       // Grade 1 reference signal
        MX_REFSIG_GRADE2 = 2,       // Grade 2 reference signal
        MX_REFSIG_RESV = 3,         // Reserved and not used until further defined.
    };
    GET_SET_FUNC(ReferenceSignal, 32, 2);
    // Sampling frequency: bit35..38
    enum MxSamplingFrequency
    {
        MX_SAMPLFREQ_NOTSET = 0,    // Not indicated (default)
        MX_SAMPLFREQ_RESV1 = 1,     // Reserved (for vectoring)
        MX_SAMPLFREQ_RESV2 = 2,     // Reserved
        MX_SAMPLFREQ_RESV3 = 3,     // Reserved
        MX_SAMPLFREQ_96000 = 4,     // 96 kHz
        MX_SAMPLFREQ_88200 = 5,     // 88,2 kHz
        MX_SAMPLFREQ_RESV4 = 6,     // Reserved
        MX_SAMPLFREQ_RESV5 = 7,     // Reserved
        MX_SAMPLFREQ_24000 = 8,     // 24 kHz
        MX_SAMPLFREQ_22050 = 9,     // 22,05 kHz
        MX_SAMPLFREQ_RESV6 = 10,    // Reserved
        MX_SAMPLFREQ_RESV7 = 11,    // Reserved
        MX_SAMPLFREQ_192000 = 12,   // 192 kHz
        MX_SAMPLFREQ_176400 = 13,   // 176,4 kHz
        MX_SAMPLFREQ_RESV8 = 14,    // Reserved
        MX_SAMPLFREQ_USRDEF = 15,   // User defined
    };
    GET_SET_FUNC(SamplingFrequency, 35, 4);
    int  GetSamplingFreqHz();
    void  SetSamplingFreqHz(int  FreqHz);
    // Sampling frequency scaling flag: bit39
    enum MxSamplingFreqScalingFlag
    {
        MX_SCALING_NONE = 0,        // No scaling (default)
        MX_SCALING_1_1001 = 1       // Sampling frequency is 1/1.001 times that indicated
                                    // either in bits 35 to 38 or bits 6 to 7
    };
    GET_SET_FUNC(SamplingFreqScalingFlag, 39, 1);

#undef  GET_SET_FUNC

    // Bit(-field) get and set helper functions
    inline unsigned char  GetBits(int  BitNr, int  NumBits);
    inline void  SetBits(unsigned char  Value, int  BitNr, int  NumBits);

    // Crc8 functions
    inline void  UpdateCrc8();
    inline void  SetCrc8(unsigned char  Crc);
    unsigned char  ComputeCrc8();
    inline  bool  IsCrc8Valid();
        
    // Data / Attributes
protected:
    unsigned char*  m_pData;
private:
    unsigned char  m_Crc8Table[256];     // CRC8 lookup for X^8+X^4+X^3+X^2+1 polynomial
        
    // Constructor / Destructor
public:
    MxAudioStatusWordParser();
    virtual ~MxAudioStatusWordParser();
};

//+=+=+=+=+=+=+=+=+=+=+=+ MxAudioStatusWordParser inline functions +=+=+=+=+=+=+=+=+=+=+=+


//-.-.-.-.-.-.-.-.-.-.-.-.-.- MxAudioStatusWordParser::GetBits -.-.-.-.-.-.-.-.-.-.-.-.-.-
//
// Get the specified bits
//   BitNr is the IEC-60958-4 bit number (Bit 0 is the first bit on the line)
//
inline unsigned char  MxAudioStatusWordParser::GetBits(int  BitNr, int  NumBits)
{
    MX_ASSERT(m_pData != NULL);
    MX_ASSERT(BitNr>=0 && BitNr<SIZE_NUM_BITS);
    MX_ASSERT(NumBits>=0 && NumBits<=8);
    MX_ASSERT(BitNr+NumBits<=SIZE_NUM_BITS);

    int  BitNrInByte = 7-(BitNr%8);
    unsigned char  Shift = (BitNrInByte+1) - NumBits;
    MX_ASSERT(Shift >= 0);

    return (m_pData[(BitNr/8)]>>Shift) & BIT_MASK[NumBits];
}
// Set the specified bits
inline void  MxAudioStatusWordParser::SetBits(
                                           unsigned char  Value, int  BitNr, int  NumBits)
{
    MX_ASSERT(m_pData != NULL);
    MX_ASSERT(BitNr>=0 && BitNr<SIZE_NUM_BITS);
    MX_ASSERT(NumBits>=0 && NumBits<=8);
    MX_ASSERT(BitNr+NumBits<=SIZE_NUM_BITS);

    int  BitNrInByte = 7-(BitNr%8);
    unsigned char  Shift = (BitNrInByte+1) - NumBits;
    MX_ASSERT(Shift >= 0);
    
    m_pData[(BitNr/8)] |= (Value & BIT_MASK[NumBits])<<Shift;
}

//-.-.-.-.-.-.-.-.-.-.-.-.-.- MxAudioStatusWordParser::SetCrc8 -.-.-.-.-.-.-.-.-.-.-.-.-.-
//
inline void  MxAudioStatusWordParser::SetCrc8(unsigned char  Crc)
{
    SetBits(Crc, 184, 8); 
}

//-.-.-.-.-.-.-.-.-.-.-.-.- MxAudioStatusWordParser::UpdateCrc8 -.-.-.-.-.-.-.-.-.-.-.-.-.
//
inline void  MxAudioStatusWordParser::UpdateCrc8()
{
    SetCrc8(ComputeCrc8());
}

//-.-.-.-.-.-.-.-.-.-.-.-.- MxAudioStatusWordParser::IsCrc8Valid -.-.-.-.-.-.-.-.-.-.-.-.-
//
inline  bool MxAudioStatusWordParser::IsCrc8Valid()
{
    MX_ASSERT(m_pData != NULL);
    unsigned char  Crc8 = ComputeCrc8();
    return (Crc8 == GetBits(184, 8));
}

//-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.- DtMxHdrMetadata -.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.
//
class DtMxHdrVpid : public DtMxAncPacket
{
    // Refer to SMPTE ST-452-1:201x, see section 4.1.6 "Payload Identifier" 
    // Refer to SMPTE ST-372:201x, see section 7 "Payload Identification" 

    // Operations
public:
    inline void SetVpid(const std::vector<unsigned char>&  Data)
    {
        m_Did = 0x41;
        m_SdidOrDbn = 0x01;
        m_Line = 8; // For NOW add to line 8

        // Add frame data
        for (int i = 0; i<(int)Data.size(); i++)
            m_FrameData.push_back(Parity(Data[i]));

        m_pUdw = m_FrameData.data();
        m_Dc = (int)m_FrameData.size();
        m_Cs = 0x00;
    }
    inline unsigned short  Parity(unsigned short  Symbol)
    {
        // Set parity bits (b8=even parity and b9=!b8)
        return PARITY_TABLE256_DATA[(Symbol & 0xFF)];
    }

    // Data / Attributes
protected:
    static const bool PARITY_TABLE256[256];     // Parity lookup table
    static const unsigned short  PARITY_TABLE256_DATA[256];     // Parity lookup table
    std::vector<unsigned short>  m_FrameData;

    // Constructor / Destructor
public:
    DtMxHdrVpid() {}
    virtual ~DtMxHdrVpid() {}
};

// Parity lookup table
const bool  DtMxHdrVpid::PARITY_TABLE256[256] = 
{
    #define P2(n) n, n^1, n^1, n
    #define P4(n) P2(n), P2(n^1), P2(n^1), P2(n)
    #define P6(n) P4(n), P4(n^1), P4(n^1), P4(n)
            P6(0), P6(1), P6(1), P6(0)
};
#undef P2
#undef P4
#undef P6

const unsigned short  DtMxHdrVpid::PARITY_TABLE256_DATA[256] =
{
    #define P0(n, v) ((((!(n))+1)<<8) | (v))
    #define P2(n, v) P0(n,v), P0(n^1,v+1), P0(n^1,v+2), P0(n,v+3)
    #define P4(n, v) P2(n,v), P2(n^1,v+4), P2(n^1,v+8), P2(n,v+12)
    #define P6(n, v) P4(n,v), P4(n^1,v+16), P4(n^1,v+32), P4(n,v+48)
            P6(0,0), P6(1,64), P6(1,128), P6(0,192)
};
#undef P0
#undef P2
#undef P4
#undef P6

class DtMxHdrMetadata : public DtMxAncPacket
{
    // Types
public:
    // Refer to SMPTE ST-2108:2018, see section 5.3.1 "Extended HDR/WCG Metadata"
    enum  FrameType
    {
        FRAME_TYPE_STATIC_TYPE1=0,        // Static Metadata Type 1
        FRAME_TYPE_ST2094_10=1,           // SMPTE ST 2094-10 HDR Dynamic Metadata Set
        FRAME_TYPE_ST2094_20=2,           // SMPTE ST 2094-20 HDR Dynamic Metadata Set
        FRAME_TYPE_ST2094_30=3,           // SMPTE ST 2094-30 HDR Dynamic Metadata Set
        FRAME_TYPE_ST2094_40=4,           // SMPTE ST 2094-40 HDR Dynamic Metadata Set
    };

    // Operations
public:
    inline unsigned short  Parity(unsigned short  Symbol)
    {
        // Set parity bits (b8=even parity and b9=!b8)
        return PARITY_TABLE256_DATA[(Symbol & 0xFF)];
    }
    inline void  CalcPacketChecksum()
    {
        int  Checksum = m_Did;
        int  DataIndex;

        Checksum += m_SdidOrDbn;
        Checksum += m_Dc;
        for (DataIndex = 0; DataIndex < m_Dc; DataIndex++)
            Checksum += m_pUdw[DataIndex];

        Checksum %= 0x200; // Checksum is sum of data (excl. header) modulo 512

        //  Bit 9 of the checksum word is then defined as the inverse of bit 8.
        if (!(Checksum & 0x100))
            Checksum |= 0x200;
        m_Cs = Checksum;
    }
    inline void  SetMetadata(const std::vector<unsigned char>&  Data)
    {
        // Set DID and SDID as defined by ST-2108:2018
        m_Did = 0x41;
        m_SdidOrDbn = 0x0C;
        m_Line = 16;    // For NOW add to line 16

        m_FrameData.clear();

        // Add frame data
        for (int i=0; i<(int)Data.size(); i++)
            m_FrameData.push_back(Parity(Data[i]));

        m_pUdw = m_FrameData.data();
        m_Dc = (int)m_FrameData.size();
        CalcPacketChecksum();
    }
protected:

    // Data / Attributes
protected:
    static const bool PARITY_TABLE256[256];     // Parity lookup table
    static const unsigned short  PARITY_TABLE256_DATA[256];     // Parity lookup table
    std::vector<unsigned short>  m_FrameData;

    // Constructor / Destructor
public:
    DtMxHdrMetadata() {}
    virtual ~DtMxHdrMetadata() {}
};

// Parity lookup table
const bool  DtMxHdrMetadata::PARITY_TABLE256[256] = 
{
    #define P2(n) n, n^1, n^1, n
    #define P4(n) P2(n), P2(n^1), P2(n^1), P2(n)
    #define P6(n) P4(n), P4(n^1), P4(n^1), P4(n)
            P6(0), P6(1), P6(1), P6(0)
};
#undef P2
#undef P4
#undef P6

const unsigned short  DtMxHdrMetadata::PARITY_TABLE256_DATA[256] =
{
    #define P0(n, v) ((((!(n))+1)<<8) | (v))
    #define P2(n, v) P0(n,v), P0(n^1,v+1), P0(n^1,v+2), P0(n,v+3)
    #define P4(n, v) P2(n,v), P2(n^1,v+4), P2(n^1,v+8), P2(n,v+12)
    #define P6(n, v) P4(n,v), P4(n^1,v+16), P4(n^1,v+32), P4(n,v+48)
            P6(0,0), P6(1,64), P6(1,128), P6(0,192)
};
#undef P0
#undef P2
#undef P4
#undef P6


} // namespace MxExamples

#endif // #ifndef  __MXHDMIHDRDEMO_H
